use mail_parser::{Addr, DateTime, HeaderValue, Message};
use std::borrow::Cow;
use std::path::{PathBuf};
use sha2::{Sha256, Digest};
use hex::encode as hex_encode;
use serde::{Deserialize, Serialize};


#[derive(Debug, Clone)]
pub struct StrMessage {
    pub id: String,
    pub hash: String,
    pub path: String,
    pub original_path: PathBuf,
    pub subject: String,
    pub thread_subject: String,
    pub received: DateTime,
    pub from: MailAddress,
    pub date: DateTime,
    pub body: String,
    pub in_reply_to: Option<String>,
    pub to: Vec<MailAddress>,
    pub cc: Vec<MailAddress>,
    pub uid: u32,
    pub list: String,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct MailAddress {
    pub name: Option<String>,
    pub address: String,
}

impl MailAddress {
    fn from_addr(addr: &Addr) -> Self {
        MailAddress {
            name: addr.name.as_ref().map(|a| a.to_string()),
            address: addr.address.as_ref().unwrap().to_string(),
        }
    }
}

impl std::fmt::Display for MailAddress {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        if let Some(n) = &self.name {
            write!(f, "\"{}\" ", n)?;
        }
        write!(f, "<{}>", self.address)
    }
}


impl StrMessage {    
    pub fn set_path(&mut self, path: String){
        self.path = path;
    }

    pub fn set_original_path(&mut self, original_path: PathBuf){
        self.original_path = original_path;
    }

    pub fn new(msg: &Message, uid: u32, list: String) -> StrMessage {
        let id = msg.message_id().unwrap_or("");
        let received = msg
            .received()
            .as_datetime_ref()
            .or_else(|| msg.date())
            .unwrap()
            .clone();
        let subject = msg.subject().unwrap_or("(No Subject)");
        let thread_subject = msg.thread_name().unwrap_or("(No Subject)");
        let invalid_email = Addr::new(None, "invalid-email");
        let from = match msg.from() {
            HeaderValue::Address(fr) => fr,
            _ => &invalid_email,
        };
        let from = MailAddress::from_addr(from);
        let date = msg.date().cloned().unwrap_or(crate::util::EPOCH);

        /// Turns a header value into a list of addresses
        fn addr_list(header: &HeaderValue) -> Vec<MailAddress> {
            match header {
                HeaderValue::Address(addr) => vec![MailAddress::from_addr(addr)],
                HeaderValue::AddressList(addrs) => {
                    addrs.iter().map(MailAddress::from_addr).collect()
                }
                _ => vec![], // TODO: should this be `unreachable!`?
            }
        }

        let to = addr_list(msg.to());
        let cc = addr_list(msg.cc());

        let in_reply_to = msg.in_reply_to().as_text_ref().map(|a| a.to_string());

        // TODO linkify body
        // TODO unformat-flowed
        let body = msg
            .body_text(0)
            .unwrap_or(Cow::Borrowed("[No message body]"));

        let mut hasher = Sha256::new();
        let data_to_encode = body.to_string().clone() + &*uid.to_string();
        hasher.update(data_to_encode);
        
        StrMessage {
            id: id.to_owned(),
            hash: hex_encode(hasher.finalize()),
            path: String::new(),
            original_path: PathBuf::new(),
            subject: subject.to_owned(),
            from,
            received,
            to,
            cc,
            thread_subject: thread_subject.to_owned(),
            date,
            body: body.to_string(),
            in_reply_to,
            uid,
            list
        }
    }
}
